#include <stdio.h>
#include <stdlib.h>
#include "btest.h"
#include <limits.h>

/* Modified 8/00 by reb for the Fall 2000 course */

/*
   Command Line options:
   -v N:  Set verbosity to level N
        0: Only give final scores
        1: Also report individual correctness scores (default)
   -f Name: Check only the named function
   -e N: Limit number of errors to report for single function to N
         (Default unbounded)
*/

static int error_limit = 1000;

static char* test_fname = NULL;

/* Generate test values near "corner cases" */
#define TEST_RANGE 3
#define TEST_COUNT 33

static int random_val(int min, int max)
{
  double weight = rand()/(double) RAND_MAX;
  int result = min * (1-weight) + max * weight;
  return result;
}

static int gen_vals(int test_vals[], int min, int max)
{
  int i;
  int test_count = 0;
  /* If range small enough, then do exhaustively */
  if (max-32 <= min) {
    for (i = min; i <= max; i++)
      test_vals[test_count++] = i;
    return test_count;
  }
  /* Otherwise, need to sample.
     Do so near the boundaries and for a few random cases */
  for (i = 0; i < TEST_RANGE; i++) {
    test_vals[test_count++] = min+i;
    test_vals[test_count++] = max-i;
    test_vals[test_count++] = (max+min)/2+i;
    test_vals[test_count++] = random_val(min, max);
  }
  return test_count;
}

static int test_0_arg(funct_t f, funct_t ft, char *name, int report)
{
  int r = f();
  int rt = ft();
  int error =  (r != rt);
  if (error && report)
    printf("Test %s() failed.\n  Gives %d[0x%x].  Should be %d[0x%x]\n",
	   name, r, r, rt, rt);
  return error;
}

static int test_1_arg(funct_t f, funct_t ft, int arg1, char *name, int report)
{
  funct1_t f1 = (funct1_t) f;
  funct1_t f1t = (funct1_t) ft;
  int r, rt, error;
  r = f1(arg1);
  rt = f1t(arg1);
  error = (r != rt);
  if (error && report)
    printf("Test %s(%d[0x%x]) failed.\n  Gives %d[0x%x].  Should be %d[0x%x]\n",
	   name, arg1, arg1, r, r, rt, rt);
  return error;
}

static int
 test_2_arg(funct_t f, funct_t ft, int arg1, int arg2, char *name, int report)
{
  funct2_t f2 = (funct2_t) f;
  funct2_t f2t = (funct2_t) ft;
  int r = f2(arg1, arg2);
  int rt = f2t(arg1, arg2);
  int error = (r != rt);
  if (error && report)
    printf(
   "Test %s(%d[0x%x],%d[0x%x]) failed.\n  Gives %d[0x%x].  Should be %d[0x%x]\n",
   name, arg1, arg1, arg2, arg2, r, r, rt, rt);
  return error;
}


static int
 test_3_arg(funct_t f, funct_t ft,
	    int arg1, int arg2, int arg3,
	    char *name, int report)
{
  funct3_t f3 = (funct3_t) f;
  funct3_t f3t = (funct3_t) ft;
  int r = f3(arg1, arg2, arg3);
  int rt = f3t(arg1, arg2, arg3);
  int error = (r != rt);
  if (error && report)
    printf(
   "Test %s(%d[0x%x],%d[0x%x],%d[0x%x]) failed.\n  Gives %d[0x%x].  Should be %d[0x%x]\n",
   name, arg1, arg1, arg2, arg2, arg3, arg3, r, r, rt, rt);
  return error;
}

/* Test a function.  Return number of errors */
static int test_function(test_ptr t, int report) {
  int test_vals[3][TEST_COUNT];
  int test_counts[3];
  int errors = 0;
  int i;
  int a1, a2, a3;
  int args = t->args;
  printf("Testing %s\n", t->name);
  /* Create test set */
  for (i = 0; i < 3; i++)
    test_counts[i] =
      gen_vals(test_vals[i], t->arg_ranges[i][0], t->arg_ranges[i][1]);
  if (args == 0) {
    errors += test_0_arg(t->solution_funct, t->test_funct,
			t->name, report && errors < error_limit);
  } else for (a1 = 0; a1 < test_counts[0]; a1++) {
    if (args == 1) {
      errors += test_1_arg(t->solution_funct, t->test_funct,
			  test_vals[0][a1],
			  t->name, report && errors < error_limit);
    } else for (a2 = 0; a2 < test_counts[1]; a2++) {
      if (args == 2) {
	errors += test_2_arg(t->solution_funct, t->test_funct,
			    test_vals[0][a1], test_vals[1][a2],
			    t->name, report && errors < error_limit);
      } else for (a3 = 0; a3 < test_counts[2]; a3++) {
	errors += test_3_arg(t->solution_funct, t->test_funct,
			    test_vals[0][a1], test_vals[1][a2],
			    test_vals[2][a3],
			    t->name, report && errors < error_limit);
      }
    }
  }
#ifndef GRADE
  if (report && errors > error_limit)
        printf("... %d total errors for function %s\n",
	       errors, t->name);
#endif
  return errors;
}

/* Run series of tests.  Return number of errors */ 
static int run_tests(int report) {
  int i;
  int errors = 0;
  double points = 0.0;
  double max_points = 0.0;
  for (i = 0; test_set[i].solution_funct; i++) {
    int terrors;
    double tscore;
    double tpoints;
    if (!test_fname || strcmp(test_set[i].name,test_fname) == 0) {
      terrors = test_function(&test_set[i], report);
#ifdef GRADE
      printf("\t%d", terrors);
#endif
      errors += terrors;
      tscore = terrors ? 0.0 : 1.0;
      tpoints = test_set[i].rating * tscore;
      points += tpoints;
      max_points += test_set[i].rating;
      if (report)
	printf("Test %s score: %.2f/%.2f\n",
	       test_set[i].name, tpoints, (double) test_set[i].rating);
    }
  }
#ifdef GRADE
  printf("\t%.2f", points);
#else
  printf("Overall Correctness score: %.2f/%.2f\n", points, max_points);
#endif
  return errors;
}



/* Command Line options:
   -v N:  Set verbosity to level N
      0: Only give final scores
      2: Also report individual correctness scores (default)
   -c: Run correctness tests only, not the performance tests
   -f Name: Check only the named function
*/

static void usage(char *cmd) {
  printf("%s [-v 0|1|2] [-c] [-f Name] [-e ELIM]\n", cmd);
  exit(1);
}

extern team_struct team;

int main(int argc, char *argv[])
{
  int i;
  int verbose_level = 1;
  int errors;
  char *cmd = argv[0];
  
#ifdef GRADE
  verbose_level = 0;
  error_limit = 0;
#endif  

  for (i = 1; i < argc; i++) {
    char* arg = argv[i];
    if (arg[0] == '-') {
      switch(arg[1]) {
        char cval;
      case 'v':
	cval = arg[2];
	if (cval == '\0') {
	  if (i < argc-1) {
	    i++;
	    arg = argv[i];
	    cval = arg[0];
	  } else
	    usage(cmd);
	}
	if (cval == '0')
	  verbose_level = 0;
	else if (cval == '1')
	  verbose_level = 1;
	else
	  usage(cmd);
	break;
      case 'f':
	test_fname = arg+2;
	if (*test_fname == '\0') {
	  if (i < argc-1) {
	    i++;
	    test_fname = argv[i];
	  } else
	    usage(cmd);
	}
	break;
      case 'e':
	/* Set error limit */
	arg+=2;
	if (*arg == '\0') {
	  if (i < argc-1) {
	    i++;
	    arg = argv[i];
	  } else
	    usage(cmd);
	}
	error_limit = atoi(arg);
	if (error_limit < 0)
	  usage(cmd);
	break;
      default:
	usage(cmd);
      }
    } else
      usage(cmd);
  }
  if(*team.name1 == '\0') {
    printf("ERROR.  Must fill in team member names!\n");
    exit(0);
  } else {
#ifdef GRADE
    printf("\t%s", team.id1);
#else
    printf("Member 1:\t%s\t%s\n", team.name1, team.id1);
#endif
  }
  if (*team.name2 != '\0') {
#ifdef GRADE
    printf("\t%s", team.id2);
#else
    printf("Member 2:\t%s\t%s\n", team.name2, team.id2);
#endif
  } else {
#ifdef GRADE
    printf("\t\"\"");
#endif
  }
  errors = run_tests(verbose_level > 0);
#ifndef GRADE
  if (errors > 0)
    printf("%d errors encountered.\n", errors);
  else {
    printf("All tests passed.\n");
  }
#endif
#ifdef GRADE
  printf("\n");
#endif
  return 0;
}
